from pathlib import Path
from subprocess import run, CalledProcessError, Popen, PIPE, STDOUT, DEVNULL
from dataclasses import dataclass
from enum import Enum

import sys
import os
import venv
import tempfile
import json
import logging
import yaml
import duckdb

from databricks.labs.lakebridge.connections.credential_manager import cred_file

from databricks.labs.lakebridge.assessments.profiler_config import PipelineConfig, Step
from databricks.labs.lakebridge.connections.database_manager import DatabaseManager, FetchResult

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

DB_NAME = "profiler_extract.db"


class StepExecutionStatus(str, Enum):
    COMPLETE = "COMPLETE"
    ERROR = "ERROR"
    SKIPPED = "SKIPPED"


@dataclass
class StepExecutionResult:
    step_name: str
    status: StepExecutionStatus
    error_message: str | None = None


class PipelineClass:
    def __init__(self, config: PipelineConfig, executor: DatabaseManager | None):
        self.config = config
        self.executor = executor
        self.db_path_prefix = Path(config.extract_folder)
        self._create_dir(self.db_path_prefix)

    def execute(self) -> list[StepExecutionResult]:
        logging.info(f"Pipeline initialized with config: {self.config.name}, version: {self.config.version}")
        execution_results: list[StepExecutionResult] = []
        error_flag = False
        for step in self.config.steps:
            logger.info(f"Executing step: {step.name}")
            result = self._process_step(step)
            execution_results.append(result)
            logger.info(f"Step '{step.name}' completed with status: {result.status}")

            # Check step execution status
            if result.status == StepExecutionStatus.ERROR:
                logger.error(f"Step {result.step_name} failed with error: {result.error_message}")
                error_flag = True
            elif result.status == StepExecutionStatus.SKIPPED:
                logger.info(f"Step {result.step_name} was skipped.")
            else:
                logger.info(f"Step {result.step_name} has completed successfully.")

        if error_flag:
            failed_steps = [r for r in execution_results if r.status == StepExecutionStatus.ERROR]
            error_msg = (
                f"Pipeline execution failed due to errors in steps: {', '.join(r.step_name for r in failed_steps)}"
            )
            logger.error(error_msg)
            raise RuntimeError(error_msg)
        return execution_results

    def _process_step(self, step: Step) -> StepExecutionResult:
        if step.flag != "active":
            logging.info(f"Skipping step: {step.name} as it is not active")
            return StepExecutionResult(step_name=step.name, status=StepExecutionStatus.SKIPPED)
        logging.debug(f"Executing step: {step.name}")
        try:
            status = self._execute_step(step)
            return StepExecutionResult(step_name=step.name, status=status)
        except RuntimeError as e:
            return StepExecutionResult(step_name=step.name, status=StepExecutionStatus.ERROR, error_message=str(e))

    def _execute_step(self, step: Step) -> StepExecutionStatus:
        if step.type == "sql":
            logging.info(f"Executing SQL step {step.name}")
            self._execute_sql_step(step)
            return StepExecutionStatus.COMPLETE
        if step.type == "python":
            logging.info(f"Executing Python step {step.name}")
            self._execute_python_step(step)
            return StepExecutionStatus.COMPLETE
        logging.error(f"Unsupported step type: {step.type}")
        raise RuntimeError(f"Unsupported step type: {step.type}")

    def _execute_sql_step(self, step: Step):
        logging.debug(f"Reading query from file: {step.extract_source}")
        with open(step.extract_source, 'r', encoding='utf-8') as file:
            query = file.read()

        if self.executor is None:
            logging.error("DatabaseManager executor is not set.")
            raise RuntimeError("DatabaseManager executor is not set.")

        # Execute the query using the database manager
        logging.info(f"Executing query: {query}")
        try:
            result = self.executor.fetch(query)

            # Save the result to duckdb
            self._save_to_db(result, step.name, str(step.mode))
        except Exception as e:
            logging.error(f"SQL execution failed: {str(e)}")
            raise RuntimeError(f"SQL execution failed: {str(e)}") from e

    def _execute_python_step(self, step: Step):

        logging.debug(f"Executing Python script: {step.extract_source}")
        db_path = str(self.db_path_prefix / DB_NAME)
        credential_config = str(cred_file("lakebridge"))
        venv_path_prefix = Path.home() / ".databricks" / "labs" / "lakebridge_profilers"
        os.makedirs(venv_path_prefix, exist_ok=True)

        # Create a temporary directory for the virtual environment
        # TODO Windows has strict checks on for temp venv cleanup, so will ignore cleanup errors and have it cleaned up later
        with tempfile.TemporaryDirectory(dir=venv_path_prefix, ignore_cleanup_errors=True) as temp_dir:
            venv_dir = Path(temp_dir) / "venv"
            venv_exec_cmd = self._create_venv(venv_dir)

            # Define the paths to the virtual environment's Python and pip executables
            if sys.platform == "win32":
                venv_python = (venv_dir / "Scripts" / "python.exe").resolve()
                venv_pip = (venv_dir / "Scripts" / "pip.exe").resolve()
            else:
                venv_python = (venv_dir / "bin" / "python").resolve()
                venv_pip = (venv_dir / "bin" / "pip").resolve()

            # Log resolved paths
            logger.info(f"Resolved venv_python: {venv_python}")
            logger.info(f"Resolved venv_pip: {venv_pip}")

            logger.info(f"Creating a virtual environment for Python script execution: {venv_dir} for step: {step.name}")
            if step.dependencies:
                self._install_dependencies(venv_exec_cmd, step.dependencies)

            self._run_python_script(venv_exec_cmd, step.extract_source, db_path, credential_config)

    @staticmethod
    def _install_dependencies(venv_exec_cmd, dependencies):
        logging.info(f"Installing dependencies: {', '.join(dependencies)}")
        try:
            logging.debug("Upgrading local pip")
            run(
                [
                    venv_exec_cmd,
                    "-m",
                    "pip",
                    "install",
                    "--upgrade",
                    "pip",
                    "--require-virtualenv",
                    "--quiet",
                    "--no-input",
                    "--disable-pip-version-check",
                ],
                check=True,
                stdout=DEVNULL,
                stderr=DEVNULL,
            )
            run(
                [
                    venv_exec_cmd,
                    "-m",
                    "pip",
                    "install",
                    *dependencies,
                    "--require-virtualenv",
                    "--quiet",
                    "--no-input",
                    "--disable-pip-version-check",
                ],
                check=True,
                stdout=DEVNULL,
                stderr=DEVNULL,
            )
        except CalledProcessError as e:
            logging.error(f"Failed to install dependencies: {e.stderr}")
            raise RuntimeError(f"Failed to install dependencies: {e.stderr}") from e

    @staticmethod
    def _run_python_script(venv_exec_cmd, script_path, db_path, credential_config):
        output_lines = []
        try:
            with Popen(
                [
                    venv_exec_cmd,
                    str(script_path),
                    "--db-path",
                    db_path,
                    "--credential-config-path",
                    credential_config,
                ],
                stdout=PIPE,
                stderr=STDOUT,
                text=True,
                bufsize=1,
            ) as process:
                if process.stdout is not None:
                    for line in process.stdout:
                        logger.info(line.rstrip())
                        output_lines.append(line)
                process.wait()
        except Exception as e:
            logging.error(f"Python script failed: {str(e)}")
            raise RuntimeError(f"Script execution failed: {str(e)}") from e

        if output_lines:
            try:
                output = json.loads(output_lines[-1])
            except json.JSONDecodeError:
                logging.info("Could not parse script output as JSON.")
                output = {
                    "status": "error",
                    "message": "Could not parse script output as JSON, manually validate the logs.",
                }

            if output.get("status") == "success":
                logging.info(f"Python script completed: {output['message']}")
            else:
                raise RuntimeError(f"Script reported error: {output.get('message', 'Unknown error')}")

        if process.returncode != 0:
            raise RuntimeError(f"Script execution failed with exit code {process.returncode}")

    def _save_to_db(self, result: FetchResult, step_name: str, mode: str):
        db_path = str(self.db_path_prefix / DB_NAME)

        with duckdb.connect(db_path) as conn:
            # TODO: Add support for figuring out data types from SQLALCHEMY result object result.cursor.description is not reliable
            schema = ' STRING, '.join(result.columns) + ' STRING'

            # Handle write modes
            if mode == 'overwrite':
                conn.execute(f"CREATE OR REPLACE TABLE {step_name} ({schema})")
            elif mode == 'append' and step_name not in conn.get_table_names(""):
                conn.execute(f"CREATE TABLE {step_name} ({schema})")

            # Batch insert using prepared statements
            placeholders = ', '.join(['?' for _ in result.columns])
            insert_query = f"INSERT INTO {step_name} VALUES ({placeholders})"

            conn.executemany(insert_query, result.rows)

    @staticmethod
    def _create_dir(dir_path: Path):
        if not Path(dir_path).exists():
            dir_path.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def load_config_from_yaml(file_path: str | Path) -> PipelineConfig:
        with open(file_path, 'r', encoding='utf-8') as file:
            data = yaml.safe_load(file)
        steps = [Step(**step) for step in data['steps']]
        return PipelineConfig(
            name=data['name'], version=data['version'], extract_folder=data['extract_folder'], steps=steps
        )

    @staticmethod
    def _create_venv(install_path: Path) -> str:
        venv_path = install_path
        # Sadly, some platform-specific variations need to be dealt with:
        #   - Windows venvs do not use symlinks, but rather copies, when populating the venv.
        #   - The library path is different.
        use_symlinks = sys.platform != "win32"

        builder = venv.EnvBuilder(with_pip=True, symlinks=use_symlinks)
        builder.create(venv_path)
        context = builder.ensure_directories(venv_path)
        logger.debug(f"Created virtual environment with context: {context}")
        return context.env_exec_cmd



import dataclasses
import logging
from abc import ABC, abstractmethod
from typing import Any
from collections.abc import Sequence, Set

from sqlalchemy import create_engine
from sqlalchemy.engine import Engine, URL
from sqlalchemy.engine.row import Row
from sqlalchemy import text
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm.session import Session

logger = logging.getLogger(__name__)
logger.setLevel("INFO")


@dataclasses.dataclass
class FetchResult:
    columns: Set[str]
    rows: Sequence[Row[Any]]


class DatabaseConnector(ABC):
    @abstractmethod
    def _connect(self) -> Engine:
        pass

    @abstractmethod
    def fetch(self, query: str) -> FetchResult:
        pass


class _BaseConnector(DatabaseConnector):
    def __init__(self, config: dict[str, Any]):
        self.config = config
        self.engine: Engine = self._connect()

    def _connect(self) -> Engine:
        raise NotImplementedError("Subclasses should implement this method")

    def fetch(self, query: str) -> FetchResult:
        if not self.engine:
            raise ConnectionError("Not connected to the database.")

        with Session(self.engine) as session, session.begin():
            result = session.execute(text(query))
            return FetchResult(result.keys(), result.fetchall())


def _create_connector(db_type: str, config: dict[str, Any]) -> DatabaseConnector:
    connectors = {
        "snowflake": SnowflakeConnector,
        "mssql": MSSQLConnector,
        "tsql": MSSQLConnector,
    }

    connector_class = connectors.get(db_type.lower())

    if connector_class is None:
        raise ValueError(f"Unsupported database type: {db_type}")

    return connector_class(config)


class SnowflakeConnector(_BaseConnector):
    def _connect(self) -> Engine:
        raise NotImplementedError("Snowflake connector not implemented")


class MSSQLConnector(_BaseConnector):
    def _connect(self) -> Engine:
        auth_type = self.config.get('auth_type', 'sql_authentication')
        db_name = self.config.get('database')

        query_params = {
            "driver": self.config['driver'],
            "loginTimeout": "30",
        }

        if auth_type == "ad_passwd_authentication":
            query_params = {
                **query_params,
                "authentication": "ActiveDirectoryPassword",
            }
        elif auth_type == "spn_authentication":
            raise NotImplementedError("SPN Authentication not implemented yet")

        connection_string = URL.create(
            drivername="mssql+pyodbc",
            username=self.config['user'],
            password=self.config['password'],
            host=self.config['server'],
            port=self.config.get('port', 1433),
            database=db_name,
            query=query_params,
        )
        return create_engine(connection_string)


class DatabaseManager:
    def __init__(self, db_type: str, config: dict[str, Any]):
        self.connector = _create_connector(db_type, config)

    def fetch(self, query: str) -> FetchResult:
        try:
            return self.connector.fetch(query)
        except OperationalError:
            logger.error("Error connecting to the database check credentials")
            raise ConnectionError("Error connecting to the database check credentials") from None

    def check_connection(self) -> bool:
        query = "SELECT 101 AS test_column"
        result = self.fetch(query)
        if result is None:
            return False
        return result.rows[0][0] == 101



import logging

from databricks.labs.blueprint.entrypoint import is_in_debug

from databricks.labs.lakebridge.cli import lakebridge
from databricks.labs.lakebridge.contexts.application import ApplicationContext

logger = logging.getLogger("databricks.labs.lakebridge.install")


def run(context: ApplicationContext):
    context.workspace_installation.uninstall(context.remorph_config)


if __name__ == "__main__":
    logger.setLevel("INFO")
    if is_in_debug():
        logging.getLogger("databricks").setLevel(logging.DEBUG)

    run(ApplicationContext(ws=lakebridge.create_workspace_client()))



import dataclasses
import tempfile
from pathlib import Path
from collections.abc import Callable

from databricks.labs.blueprint.entrypoint import get_logger
from databricks.labs.blueprint.tui import Prompts

from databricks.labs.bladespector.analyzer import Analyzer

from databricks.labs.lakebridge.helpers.file_utils import check_path, move_tmp_file

logger = get_logger(__file__)


@dataclasses.dataclass
class AnalyzerResult:
    source_directory: Path
    output_directory: Path
    source_system: str


class AnalyzerPrompts:

    def __init__(self, prompts: Prompts):
        self._prompts = prompts

    def get_source_directory(self) -> Path:
        """Get and validate the source directory from user input."""
        directory_str = self._prompts.question(
            "Enter full path to the source directory",
            default=Path.cwd().as_posix(),
            validate=check_path,
        )
        return Path(directory_str).resolve()

    def get_result_file_path(self, directory: Path) -> Path:
        """Get the result file path - accepts either filename or full path."""
        filename = self._prompts.question(
            "Enter report file name or custom export path including file name without extension",
            default=f"{directory.as_posix()}/lakebridge-analyzer-results.xlsx",
            validate=check_path,
        )
        return directory / Path(filename) if len(filename.split("/")) == 1 else Path(filename)

    def get_source_system(self, platform: str | None = None) -> str:
        """Validate source technology or prompt for a valid source"""
        if platform is None or platform not in Analyzer.supported_source_technologies():
            if platform is not None:
                logger.warning(f"Invalid source technology {platform}")
            platform = self._prompts.choice("Select the source technology", Analyzer.supported_source_technologies())
        assert platform in Analyzer.supported_source_technologies()

        return platform


class AnalyzerRunner:
    def __init__(
        self, runnable: Callable[[Path, Path, str, bool], None], move_file: Callable[[Path, Path], None], is_debug: bool
    ):
        self._runnable = runnable
        self._move_file = move_file
        self._is_debug = is_debug

    @classmethod
    def create(cls, is_debug: bool = False) -> "AnalyzerRunner":
        return cls(Analyzer.analyze, move_tmp_file, is_debug)

    def run(self, source_dir: Path, results_dir: Path, platform: str) -> AnalyzerResult:
        logger.debug(f"Starting analyzer execution in ${source_dir} for ${platform}")

        if not check_path(source_dir) or not check_path(results_dir):
            raise ValueError(f"Invalid path(s) provided: source_dir={source_dir}, results_dir={results_dir}")

        tmp_dir = self._temp_xlsx_path(results_dir)
        self._runnable(source_dir, tmp_dir, platform, self._is_debug)
        self._move_file(tmp_dir, Path(results_dir))
        logger.info(f"Successfully Analyzed files in ${source_dir} for ${platform} and saved report to {results_dir}")
        return AnalyzerResult(Path(source_dir), Path(results_dir), platform)

    @staticmethod
    def _temp_xlsx_path(results_dir: Path | str) -> Path:
        return (Path(tempfile.mkdtemp()) / Path(results_dir).name).with_suffix(".xlsx")


class LakebridgeAnalyzer:

    def __init__(self, prompts: AnalyzerPrompts, runner: AnalyzerRunner):
        self._prompts = prompts
        self._runner = runner

    def run_analyzer(
        self, source: str | None = None, results: str | None = None, platform: str | None = None
    ) -> AnalyzerResult:
        if not source:
            source_dir = self._prompts.get_source_directory()
        elif not isinstance(source, Path):
            source_dir = Path(source)
        else:
            source_dir = source

        if not results:
            results_dir = self._prompts.get_result_file_path(source_dir)
        elif not isinstance(results, Path):
            results_dir = Path(results)
        else:
            results_dir = results

        platform = self._prompts.get_source_system(platform)

        return self._runner.run(source_dir, results_dir, platform)



import pandas as pd
from azure.monitor.query import MetricAggregationType
from datetime import datetime, timedelta, time as timex
from azure.synapse.artifacts import models as ArtifactsModels

from abc import ABC


# AzureArtifact
class AzureArtifact(ABC):
    # constructor
    def __init__(self, tz_info, artifacts_client, fetch_batch_size=20, max_pages=5000):
        self.tz_info = tz_info
        self.client = artifacts_client
        self.fetch_batch_size = fetch_batch_size
        self.max_pages = max_pages

    @staticmethod
    def project_dict(obj, keep=[], remove=[]):
        """
        Utility Function to keep or/and remove fields
        """
        return {k: v for (k, v) in obj.items() if (not keep or k.lower() in keep) and (k.lower() not in remove)}

    @staticmethod
    def create_run_filter_parameters(last_updated_after, last_updated_before):
        # TODO: fix this from global context to import
        return ArtifactsModels.RunFilterParameters(
            last_updated_after=last_updated_after, last_updated_before=last_updated_before
        )

    # fetch_from_iter
    def fetch_from_iter(self, iterator, keep, remove):
        """
        Creates item groups ( lists of max size fetch_batch_size) from input iterator
        """
        group = []
        for entry in iterator:
            # add item to group
            group.append(AzureArtifact.project_dict(entry.as_dict(), keep, remove))
            # yield the group list for every batch size
            if len(group) >= self.fetch_batch_size:
                yield group
                # make sure to clear the group after yield
                group.clear()
        # make sure to emit eny partial groups after looping
        if len(group) > 0:
            yield group

    # query_activity_runs
    def query_activity_runs(self, runs_query, run_filter_parameters, keep, remove):
        """
        Creates item groups ( lists of max size fetch_batch_size) from the input query after executing it using run_filter_parameters
        query response here has pagination logic. It should have following fields
          i)  value : list of items (current page)
          ii) continuation_token: The continuation token for getting the next page of results, if any remaining results exist, null otherwise.

          Example Model:
            https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.models.pipelinerunsqueryresponse?view=azure-python-preview
        """
        group = []
        page_count = 0
        run_filter_parameters.continuation_token = None
        while page_count == 0 or (run_filter_parameters.continuation_token and page_count <= self.max_pages):
            page_result = runs_query(filter_parameters=run_filter_parameters)
            page_count += 1
            for entry in page_result.value:
                # add item to group
                group.append(AzureArtifact.project_dict(entry.as_dict(), keep, remove))
                # yield the group list for every batch size
                if len(group) >= self.fetch_batch_size:
                    yield group
                    # make sure to clear the group after yield
                    group.clear()
            # update continuation_token for run_filter_parameters
            run_filter_parameters.continuation_token = page_result.continuation_token
        # make sure to emit eny partial groups after outer looping
        if len(group) > 0:
            yield group


"""
  SynapseWorkspace
"""


class SynapseWorkspace(AzureArtifact):
    """
    constructor

    :artifacts_client - create one using *get_synapse_artifacts_client* utility function

    Refereces:
      - https://learn.microsoft.com/en-us/python/api/azure-synapse/azure.synapse?view=azure-python-preview

    """

    # constructor
    def __init__(self, tz_info, artifacts_client, fetch_batch_size=20):
        super().__init__(tz_info, artifacts_client, fetch_batch_size)

    # get_workspace_info
    def get_workspace_info(
        self,
        keep=[
            'id',
            'name',
            'type',
            'workspace_uid',
            'location',
            'provisioning_state',
            'default_data_lake_storage',
            'workspace_repository_configuration',
            'purview_configuration',
            'extra_properties',
        ],
        remove=[],
    ):
        """
        Query workspace info
        """
        workspace = self.client.workspace.get()
        return AzureArtifact.project_dict(workspace.as_dict(), keep, remove)

    # list_sql_pools
    def list_sql_pools(
        self, keep=['id', 'name', 'type', 'location', 'sku', 'provisioning_state', 'status', 'creation_date'], remove=[]
    ):
        """
        Query SQL Pools
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.sqlpoolsoperations?view=azure-python-preview#azure-synapse-artifacts-operations-sqlpoolsoperations-list
        """
        result = self.client.sql_pools.list()  # returns SqlPoolInfoListResult
        yield from self.fetch_from_iter(result.value, keep, remove)

    # list_bigdata_pools
    def list_bigdata_pools(self, keep=[], remove=[]):
        """
        Query Spark Pools
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.bigdatapoolsoperations?view=azure-python-preview#azure-synapse-artifacts-operations-bigdatapoolsoperations-list
        """
        result = self.client.big_data_pools.list()  # returns BigDataPoolResourceInfoListResult
        yield from self.fetch_from_iter(result.value, keep, remove)

    # list_linked_services
    def list_linked_services(self, keep=[], remove=[]):
        """
        Query Pipe Lines
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.linkedserviceoperations?view=azure-python-preview#azure-synapse-artifacts-operations-linkedserviceoperations-get-linked-services-by-workspace
        """
        result = (
            self.client.linked_service.get_linked_services_by_workspace()
        )  # returns ItemPaged[LinkedServiceResource]
        yield from self.fetch_from_iter(result, keep, remove)

    # list_data_flows
    def list_data_flows(self, keep=[], remove=[]):
        """
        Query Data Flows
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.dataflowoperations?view=azure-python-preview#azure-synapse-artifacts-operations-dataflowoperations-get-data-flows-by-workspace

        """
        result = self.client.data_flow.get_data_flows_by_workspace()  # returns ItemPaged[DataFlowResource]
        yield from self.fetch_from_iter(result, keep, remove)

    # list_pipelines
    def list_pipelines(self, keep=[], remove=[]):
        """
        Query Pipelines
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.pipelineoperations?view=azure-python-preview#azure-synapse-artifacts-operations-pipelineoperations-get-pipelines-by-workspace
        """
        result = self.client.pipeline.get_pipelines_by_workspace()  # returns ItemPaged[PipelineResource]
        yield from self.fetch_from_iter(result, keep, remove)

    # list_notebooks
    def list_notebooks(self, keep=[], remove=[]):
        """
        Query Noteboos
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.notebookoperations?view=azure-python-preview#azure-synapse-artifacts-operations-notebookoperations-get-notebooks-by-workspace
        """
        result = self.client.notebook.get_notebooks_by_workspace()  # returns ItemPaged[NotebookResource]
        yield from self.fetch_from_iter(result, keep, remove)

    # list_spark_job_definitions
    def list_spark_job_definitions(self, keep=[], remove=[]):
        """
        Query Spark Jobs
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.sparkjobdefinitionoperations?view=azure-python-preview#azure-synapse-artifacts-operations-sparkjobdefinitionoperations-get-spark-job-definitions-by-workspace
        """
        result = (
            self.client.spark_job_definition.get_spark_job_definitions_by_workspace()
        )  # returns ItemPaged[SparkJobDefinitionResource]
        yield from self.fetch_from_iter(result, keep, remove)

    # list_sqlscripts
    def list_sqlscripts(self, keep=[], remove=[]):
        """
        Query Pipe Lines
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.sqlscriptoperations?view=azure-python-preview#azure-synapse-artifacts-operations-sqlscriptoperations-get-sql-scripts-by-workspace
        """
        result = self.client.sql_script.get_sql_scripts_by_workspace()  # rerurns ItemPaged[SqlScriptResource]
        yield from self.fetch_from_iter(result, keep, remove)

        # list_triggers

    def list_triggers(self, keep=[], remove=[]):
        """
        Query triggers
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.triggeroperations?view=azure-python-preview#azure-synapse-artifacts-operations-triggeroperations-get-triggers-by-workspace
        """
        result = self.client.trigger.get_triggers_by_workspace()  # returns ItemPaged[TriggerResource]
        yield from self.fetch_from_iter(result, keep, remove)

        # list_libraries

    def list_libraries(self, keep=[], remove=[]):
        """
        Query Libraries
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.libraryoperations?view=azure-python-preview#azure-synapse-artifacts-operations-libraryoperations-list
        """
        result = self.client.library.list()  # returns ItemPaged[LibraryResource]
        yield from self.fetch_from_iter(result, keep, remove)

        # list_datasets

    def list_datasets(self, keep=[], remove=[]):
        """
        Query Pipe Lines
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.datasetoperations?view=azure-python-preview#azure-synapse-artifacts-operations-datasetoperations-get-datasets-by-workspace
        """
        result = self.client.dataset.get_datasets_by_workspace()  # returns ItemPaged[DatasetResource]
        yield from self.fetch_from_iter(result, keep, remove)

    # list_pipeline_runs_dep
    def list_pipeline_runs_dep(self, run_filter_params, keep=[], remove=[]):
        """
        Query Pipe Lines
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.pipelinerunoperations?view=azure-python-preview#azure-synapse-artifacts-operations-pipelinerunoperations-query-pipeline-runs-by-workspace
        """
        result = self.client.pipeline_run.query_pipeline_runs_by_workspace(
            filter_parameters=run_filter_params
        )  # returns PipelineRunsQueryResponse

        for run in result.value:
            yield AzureArtifact.project_dict(run.as_dict(), keep, remove)

    # list_pipeline_runs
    def list_pipeline_runs(self, last_updated_date, keep=[], remove=[]):
        """
        Query Pipeline runs by last_updted_date (in UTC)
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.pipelinerunoperations?view=azure-python-preview#azure-synapse-artifacts-operations-pipelinerunoperations-query-pipeline-runs-by-workspace
        """
        last_updated_after = datetime.combine(last_updated_date, timex(0, 0, 0, 0)).replace(tzinfo=self.tz_info)
        last_updated_before = datetime.combine(last_updated_date, timex(23, 59, 59, 999999)).replace(
            tzinfo=self.tz_info
        )

        # set RunFilterParameters
        run_filter_params = AzureArtifact.create_run_filter_parameters(last_updated_after, last_updated_before)
        runs_query = self.client.pipeline_run.query_pipeline_runs_by_workspace
        yield from self.query_activity_runs(runs_query, run_filter_params, keep, remove)

    # list_trigger_runs
    def list_trigger_runs(self, last_updated_date, keep=[], remove=[]):
        """
        Query Pipe Lines
        SDK Reference:
          https://learn.microsoft.com/en-us/python/api/azure-synapse-artifacts/azure.synapse.artifacts.operations.triggerrunoperations?view=azure-python-preview#azure-synapse-artifacts-operations-triggerrunoperations-query-trigger-runs-by-workspace
        """
        last_updated_after = datetime.combine(last_updated_date, timex(0, 0, 0, 0)).replace(tzinfo=self.tz_info)
        last_updated_before = datetime.combine(last_updated_date, timex(23, 59, 59, 999999)).replace(
            tzinfo=self.tz_info
        )

        # set RunFilterParameters
        run_filter_params = AzureArtifact.create_run_filter_parameters(last_updated_after, last_updated_before)
        runs_query = self.client.trigger_run.query_trigger_runs_by_workspace
        yield from self.query_activity_runs(runs_query, run_filter_params, keep, remove)


# SynapseMetrics
class SynapseMetrics:
    def __init__(self, metrics_client, num_days=90, granularity_mins=15, fetch_batch_size=500, max_pages=5000):
        self.client = metrics_client
        self.num_days = num_days
        self.granularity_mins = granularity_mins
        self.fetch_batch_size = fetch_batch_size
        self.max_pages = max_pages

    # fetch_metrics
    def fetch_metrics(self, metrics):
        """
        Creates item groups ( lists of max size fetch_batch_size) from input iterator
        """
        data = []
        for metric in metrics:
            for ts_entry in metric.timeseries:
                for metric_value in ts_entry.data:
                    # add item to group
                    data.append(
                        {
                            "name": metric.name,
                            "timestamp": metric_value.timestamp,
                            "average": metric_value.average,
                            "count": metric_value.count,
                            "maximum": metric_value.maximum,
                            "minimum": metric_value.minimum,
                            "total": metric_value.total,
                        }
                    )
        return pd.DataFrame(data)

    # get_dedicated_pool_metrics
    def get_dedicated_sql_pool_metrics(self, resource_id):
        """
        Quries metrics for a specific dedicated sql metric
        resource_id: input sql pool resource id
        """
        response = self.client.query_resource(
            resource_id,
            metric_names=[
                "DWULimit",
                "DWUUsed",
                "DWUUsedPercent",
                "MemoryUsedPercent",
                "CPUPercent",
                "Connections",
                "ActiveQueries",
            ],
            timespan=timedelta(days=self.num_days),
            granularity=timedelta(minutes=self.granularity_mins),
            aggregations=[
                MetricAggregationType.AVERAGE,
                MetricAggregationType.COUNT,
                MetricAggregationType.MINIMUM,
                MetricAggregationType.MAXIMUM,
                MetricAggregationType.TOTAL,
            ],
        )
        # Fetch Metrics
        return self.fetch_metrics(response.metrics)

    # get_spark_pool_metrics
    def get_spark_pool_metrics(self, resource_id):
        """
        Query metrics for a specific spark pool
        resource_id: input spark pool resource id
        """
        response = self.client.query_resource(
            resource_id,
            metric_names=[
                "BigDataPoolApplicationsEnded",
                "BigDataPoolAllocatedCores",
                "BigDataPoolAllocatedMemory",
                "BigDataPoolApplicationsActive",
            ],
            timespan=timedelta(days=self.num_days),
            granularity=timedelta(minutes=self.granularity_mins),
            aggregations=[
                MetricAggregationType.AVERAGE,
                MetricAggregationType.COUNT,
                MetricAggregationType.MINIMUM,
                MetricAggregationType.MAXIMUM,
                MetricAggregationType.TOTAL,
            ],
        )
        # Fetch Metrics
        return self.fetch_metrics(response.metrics)

    # get_workspace_level_metrics
    def get_workspace_level_metrics(self, resource_id):
        """
        Query Workspace level metrics
        resource_id: input workspace resource id
        """
        response = self.client.query_resource(
            resource_id,
            metric_names=[
                "IntegrationActivityRunsEnded",
                "IntegrationPipelineRunsEnded",
                "IntegrationTriggerRunsEnded",
                "BuiltinSqlPoolDataProcessedBytes",
                "BuiltinSqlPoolLoginAttempts",
                "BuiltinSqlPoolRequestsEnded",
            ],
            timespan=timedelta(days=self.num_days),
            granularity=timedelta(hours=1),
            aggregations=[
                MetricAggregationType.AVERAGE,
                MetricAggregationType.COUNT,
                MetricAggregationType.MINIMUM,
                MetricAggregationType.MAXIMUM,
                MetricAggregationType.TOTAL,
            ],
        )
        # Fetch Metrics and return as DataFrame
        return self.fetch_metrics(response.metrics)



from databricks.labs.lakebridge.connections.database_manager import DatabaseManager


def get_sqlpool_reader(
    input_cred: dict,
    db_name: str,
    *,
    endpoint_key: str = 'dedicated_sql_endpoint',
    auth_type: str = 'sql_authentication',
) -> DatabaseManager:
    config = {
        "driver": input_cred['driver'],
        "server": input_cred[endpoint_key],
        "database": db_name,
        "user": input_cred['sql_user'],
        "password": input_cred['sql_password'],
        "port": input_cred.get('port', 1433),
        "auth_type": auth_type,
    }
    # synapse and mssql use the same connector
    source = "mssql"

    return DatabaseManager(source, config)
